import logging
import types
import warnings

from . import _loggers
from . import _settings

log = _loggers.get_logger(__name__)


def _patch_loggers(settings: _settings.Base) -> None:
    # root logger is not in `loggerDict` so we have to patch it separately
    _configure_logger(settings, logging.root)
    for logger in logging.Logger.manager.loggerDict.values():
        _configure_logger(settings, logger)


def _configure_logger(
    settings: _settings.Base, logger: logging.Logger | logging.PlaceHolder
) -> None:
    if isinstance(logger, (_loggers.Logtator, logging.PlaceHolder)):
        return

    logger.makeRecord = types.MethodType(_loggers.Logtator.makeRecord, logger)  # type: ignore[method-assign]
    logger._log = types.MethodType(_loggers.Logtator._log, logger)  # type: ignore[method-assign]
    logger.addHandler = types.MethodType(_loggers.Logtator.addHandler, logger)  # type: ignore[method-assign]
    logger.bind = types.MethodType(_loggers.Logtator.bind, logger)  # type: ignore[attr-defined]

    logger.handlers = []
    logger.propagate = True

    if settings.force_default_level:
        logger.setLevel(settings.default_level)


def _patch_warning(settings: _settings.Base) -> None:
    warnings.showwarning = types.MethodType(
        _loggers.Logtator._log_warning, logging.root
    )
    if settings.reset_warning_filters:
        warnings.filters = []


def patch(settings: _settings.Base) -> None:
    if logging.getLoggerClass() is _loggers.Logtator:
        log.debug("already-patched")
        return

    logging.setLoggerClass(_loggers.Logtator)
    logging.root.setLevel(settings.default_level)
    logging.lastResort.setLevel(settings.default_level)  # type: ignore[union-attr]
    logging._defaultFormatter = settings.logs_formatter  # type: ignore[attr-defined]

    _patch_warning(settings)
    _patch_loggers(settings)

    log.info(
        "known-loggers",
        loggers=[
            {
                "name": logger.name,
                "level": logger.level,
                "propagate": logger.propagate,
                "disabled": logger.disabled,
                "handlers": logger.handlers,
            }
            if log.level == logging.DEBUG
            else logger.name
            for logger in logging.Logger.manager.loggerDict.values()
            if not isinstance(logger, logging.PlaceHolder)
        ],
    )



import logging
import typing


class Logger(logging.Logger):
    def debug(self, msg: object, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    def info(self, msg: object, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    def warning(self, msg: object, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    def error(self, msg: object, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    def exception(
        self, msg: object, *args: typing.Any, **kwargs: typing.Any
    ) -> None: ...

    def critical(
        self, msg: object, *args: typing.Any, **kwargs: typing.Any
    ) -> None: ...

    def log(
        self, level: int, msg: object, *args: typing.Any, **kwargs: typing.Any
    ) -> None: ...

    def bind(self, **kwargs: typing.Any) -> "Logger": ...  # type: ignore[empty-body]



__all__ = [
    "annotations",
    "encoders",
    "formatters",
    "get_logger",
    "getLogger",
    "patch",
    "Settings",
]

from ._base import patch
from ._loggers import get_logger, getLogger
from ._settings import Base as Settings



import src.logtator as ld


def test_patch(capsys):
    settings = ld.Settings()
    ld.patch(settings)

    import logging
    log = logging.getLogger()
    # msg, *args, **kwargs
    log.warning("Hello, world!", kol="er")
    log.warning("Houston, we have a %s", "interesting problem")
    captured = capsys.readouterr()
    assert captured.err == "Hello, world!\nHouston, we have a interesting problem\n"




import dataclasses
import logging

from . import formatters


@dataclasses.dataclass
class Base:
    default_level: int | str = logging.INFO
    force_default_level: bool = False
    logs_formatter: logging.Formatter = formatters.Json()
    reset_warning_filters: bool = True



import datetime as dt
import json
import logging
import traceback
import typing

from . import encoders


class HumanReadable(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        if record.exc_info:
            traceback.print_exception(*record.exc_info)
        return (
            f"{record.levelname}: {record.getMessage()} {record.extra or ''} "  # type: ignore[attr-defined]
            f"{record.kwargs or ''}"  # type: ignore[attr-defined]
        ).strip()


class Json(logging.Formatter):
    def __init__(
        self,
        fmt: str | None = None,
        datefmt: str | None = None,
        style: typing.Literal["%", "{", "$"] = "%",
        validate: bool = True,
        *,
        defaults: typing.Dict[str, typing.Any] | None = None,
        flatten_extra: bool = True,
        flatten_kwargs: bool = True,
        nest_record: bool = True,
        add_google_cloud_fields: bool = False,
    ) -> None:
        super().__init__(
            fmt=fmt, datefmt=datefmt, style=style, validate=validate, defaults=defaults
        )
        self.flatten_extra = flatten_extra
        self.flatten_kwargs = flatten_kwargs
        self.nest_record = nest_record
        # https://cloud.google.com/logging/docs/agent/logging/configuration#special-fields
        self.add_google_cloud_fields = add_google_cloud_fields

    def format(self, record: logging.LogRecord) -> str:
        record_dict = record.__dict__
        if self.nest_record:
            record_dict = {"record": record.__dict__}
        record_dict["message"] = record.getMessage()

        if self.flatten_extra:
            record_dict.update(
                record_dict.get("record", record_dict).pop("extra", {}) or {}
            )
        if self.flatten_kwargs:
            record_dict.update(
                record_dict.get("record", record_dict).pop("kwargs", {}) or {}
            )

        if self.add_google_cloud_fields:
            map_google_severity = {
                "NOTSET": "DEFAULT",
                "WARN": "WARNING",
                "FATAL": "CRITICAL",
            }
            record_dict["severity"] = map_google_severity.get(
                record.levelname, record.levelname
            )
            record_dict["time"] = (
                dt.datetime.fromtimestamp(
                    record.__dict__["created"], tz=dt.timezone.utc
                )
                .isoformat()
                .replace("+00:00", "Z")
            )
        return json.dumps(record_dict, cls=encoders.Base)



import datetime
import json
import traceback
import types
import typing


class Base(json.JSONEncoder):
    def default(self, obj: object) -> str | typing.List[str] | typing.Any:
        if isinstance(obj, datetime.date):
            return obj.isoformat()
        if isinstance(obj, types.TracebackType):
            return traceback.format_tb(obj)

        try:
            return super().default(obj)
        except TypeError:
            return str(obj)

    def _recursive_convert_key(self, obj: object) -> dict[str, object] | object:
        if isinstance(obj, dict):
            return {
                self.default(key): self._recursive_convert_key(value)
                for key, value in obj.items()
            }
        return obj

    def encode(self, obj: object) -> str:
        try:
            return super().encode(obj)
        except TypeError:
            return super().encode(self._recursive_convert_key(obj))



import linecache
import logging
import sys
import types
import typing

from . import annotations


def get_logger(name: str) -> annotations.Logger:
    return logging.getLogger(name)  # type: ignore[return-value]


getLogger = get_logger
log = get_logger(__name__)


class Logtator(logging.Logger):
    def makeRecord(
        self,
        name: str,
        level: int,
        fn: str,
        lno: int,
        msg: object,
        args: typing.Tuple[object, ...] | typing.Mapping[str, object],
        exc_info: tuple[type[BaseException], BaseException, types.TracebackType | None]
        | tuple[None, None, None]
        | None,
        func: str | None = None,
        extra: typing.Mapping[str, object] | None = None,
        sinfo: str | None = None,
        **kwargs: typing.Any,
    ) -> logging.LogRecord:
        rv = logging._logRecordFactory(  # type: ignore[attr-defined]
            name, level, fn, lno, msg, args, exc_info, func, sinfo
        )
        rv.extra = extra
        rv.kwargs = kwargs
        return rv  # type: ignore[no-any-return]

    def _log(
        self,
        level: int,
        msg: object,
        args: typing.Tuple[object, ...] | typing.Mapping[str, object],
        exc_info: bool
        | tuple[type[BaseException], BaseException, types.TracebackType | None]
        | tuple[None, None, None]
        | BaseException
        | None
        | None = None,
        extra: typing.Mapping[str, object] | None = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        **kwargs: typing.Any,
    ) -> None:
        sinfo = None
        if logging._srcfile:
            try:
                # We need to increase `stacklevel` by 1 otherwise caller would be this
                # line instead of the actual caller.
                fn, lno, func, sinfo = self.findCaller(stack_info, stacklevel + 1)
            except ValueError:
                fn, lno, func = "(unknown file)", 0, "(unknown function)"
        else:
            fn, lno, func = "(unknown file)", 0, "(unknown function)"
        if exc_info:
            if isinstance(exc_info, BaseException):
                exc_info = (type(exc_info), exc_info, exc_info.__traceback__)
            elif not isinstance(exc_info, tuple):
                exc_info = sys.exc_info()
        record = self.makeRecord(
            self.name,
            level,
            fn,
            lno,
            msg,
            args,
            exc_info,  # type: ignore[arg-type]
            func,
            extra,
            sinfo,
            **kwargs,
        )
        self.handle(record)

    def _log_warning(
        self,
        message: str,
        category: str,
        filename: str,
        lineno: int,
        file: str | None = None,
        line: str | None = None,
    ) -> None:
        if file:
            log.debug("warning-file-ignored", file=file)
        if not line:
            line = linecache.getline(filename, lineno)
        record = self.makeRecord(
            self.name,
            logging.WARNING,
            filename,
            lineno,
            message,
            (),
            None,
            category=category,
            location=f"{filename}:{lineno}",
            line=line,
        )
        self.handle(record)

    def addHandler(self, handler: logging.Handler) -> None:
        log.debug(
            "handler-thrown-away",
            logger=self.name,
            handler=handler.name,
        )

    def bind(self, **kwargs: typing.Any) -> "_Adapter":
        return _Adapter(self, **kwargs)


class _Adapter(logging.LoggerAdapter[Logtator]):
    def __init__(self, logger: Logtator, **kwargs: typing.Any) -> None:
        super().__init__(logger)
        self.kwargs = kwargs

    def process(
        self, msg: str, kwargs: typing.Any
    ) -> typing.Tuple[str, typing.MutableMapping[str, typing.Any]]:
        kwargs.update(self.kwargs)
        return msg, kwargs
